
# https://cmake.org/pipermail/cmake/2007-March/013081.html
set(CMAKE_CONFIGURATION_TYPES Debug Release CACHE TYPE INTERNAL FORCE)

STRING(REGEX REPLACE ".*\/(.+)\/.+$" "\\1" CAT_DIR ${CMAKE_BINARY_DIR})
IF (NOT CAT_DIR)
	MESSAGE(FATAL_ERROR "cannot set CAT_DIR")
ENDIF()

IF(${NLIB_DLL})
	OPTION(BUILD_SHARED_LIBS "Build shared libraries (DLLs)." ON)
	SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_DEFINITIONS "NLIB_WINDLL")
ELSE()
	OPTION(BUILD_SHARED_LIBS "Build shared libraries (DLLs)." OFF)
ENDIF()

# VS2013 or above
IF(MSVC_VERSION VERSION_LESS 1800)
	MESSAGE(FATAL_ERROR "!!!!! Please use Visual Studio 2013 or later. !!!!!")
ENDIF()

# Only 64bit Windows
SET(ProgramW6432 "$ENV{ProgramW6432}")
IF(NOT ProgramW6432)
	MESSAGE(FATAL_ERROR "!!!!! Please use 64bit Windows !!!!!") 
ENDIF()

# From cmake 3.8, you can install the binary version of boost into C:\boost
# https://sourceforge.net/projects/boost/files/boost-binaries/
IF(CMAKE_SIZEOF_VOID_P EQUAL 8)
	# PROGRAMFILES has to be chagned if cmake is invoked from 32bit shell.
	SET(ENV{PROGRAMFILES} "$ENV{ProgramW6432}")
	IF(CMAKE_VERSION VERSION_LESS "3.8.0")
		IF(NOT DEFINED BOOSTROOT)
			SET(BOOSTROOT "D:/Program Files/boost") # workaround for D-drive users (use junction)
		ENDIF()
	ENDIF()
ELSE()
	# PROGRAMFILES has to be chagned if cmake is invoked from 64bit shell.
	SET(ENV{PROGRAMFILES} "$ENV{ProgramFiles\(x86\)}")
	IF(CMAKE_VERSION VERSION_LESS "3.8.0")
		IF(NOT DEFINED BOOSTROOT)
			SET(BOOSTROOT "D:/Program Files (x86)/boost") # workaround for D-drive users (use junction)
		ENDIF()
	ENDIF()
ENDIF()

IF(NOT DEFINED ZLIB_ROOT)
	SET(ZLIB_ROOT "$ENV{PROGRAMFILES}/zlib")
ENDIF()

# SET(Boost_DEBUG TRUE)
IF(CMAKE_VERSION VERSION_LESS "3.8.0")
	IF(NOT DEFINED BOOST_ROOT)
		SET(BOOST_ROOT "$ENV{PROGRAMFILES}/boost")
	ENDIF()
ENDIF()

LIST(APPEND CMAKE_INCLUDE_PATH "$ENV{PROGRAMFILES}/curl/include")
LIST(APPEND CMAKE_LIBRARY_PATH "$ENV{PROGRAMFILES}/curl/lib")

SET(CMAKE_DEBUG_POSTFIX "-${CAT_DIR}_d")
SET(CMAKE_RELEASE_POSTFIX "-${CAT_DIR}")

#
# Compiler options
#
# http://www.cmake.org/Wiki/CMake_FAQ#Dynamic_Replace
FOREACH(flag_var
         CMAKE_CXX_FLAGS CMAKE_CXX_FLAGS_DEBUG CMAKE_CXX_FLAGS_RELEASE
         CMAKE_CXX_FLAGS_MINSIZEREL CMAKE_CXX_FLAGS_RELWITHDEBINFO
         CMAKE_C_FLAGS CMAKE_C_FLAGS_DEBUG CMAKE_C_FLAGS_RELEASE
         CMAKE_C_FLAGS_MINSIZEREL CMAKE_C_FLAGS_RELWITHDEBINFO)
	STRING(REPLACE "/Zi" "/Z7" ${flag_var} "${${flag_var}}")
	STRING(REPLACE "/W3" "/W4" ${flag_var} "${${flag_var}}")
ENDFOREACH()
SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_DEFINITIONS "NOMINMAX;WIN32_LEAN_AND_MEAN;_SCL_SECURE_NO_WARNINGS;_USE_MATH_DEFINES")

IF(MSVC_VERSION EQUAL 1800)
	SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "/Zc:rvalueCast")
ENDIF()
IF(NOT CMAKE_CL_64)
	SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "/arch:SSE2")
ENDIF()
SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "/volatile:iso")
SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "/bigobj;/Oi")
SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "/sdl")
SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "$<$<CONFIG:Debug>:/Ob1>")
IF(MSVC_VERSION EQUAL 1900)
	# disable C4592 temporarily for VS2015 Update1
	# see https://stackoverflow.com/questions/34013930/error-c4592-symbol-will-be-dynamically-initialized-vs2015-1-static-const-std
	SET_PROPERTY(DIRECTORY APPEND PROPERTY COMPILE_OPTIONS "/wd4592")
ENDIF()

# To remove _MBCS
ADD_DEFINITIONS(-D_SBCS)

MACRO(library_default_build_rule target)
	IF(${NLIB_DLL})
		IF(EXISTS ${NLIB_PKGROOT_DIR}/../_common/cmake/version.rc.in)
			configure_file(
				${NLIB_PKGROOT_DIR}/../_common/cmake/version.rc.in
				${CMAKE_CURRENT_BINARY_DIR}/version.rc
				@ONLY)
			ADD_LIBRARY(${target} SHARED ${LIB_SRC} ${LIB_HEADER} ${CMAKE_CURRENT_BINARY_DIR}/version.rc)
			IF(ZLIB_FOUND)
				TARGET_LINK_LIBRARIES(${target} ${ZLIB_LIBRARIES})
			ENDIF()
		ELSE()
			ADD_LIBRARY(${target} STATIC ${LIB_SRC} ${LIB_HEADER})
		ENDIF()
	ELSE()
		ADD_LIBRARY(${target} STATIC ${LIB_SRC} ${LIB_HEADER})
	ENDIF()
	SET_PROPERTY(TARGET ${target} APPEND PROPERTY INTERFACE_INCLUDE_DIRECTORIES "$<BUILD_INTERFACE:${NLIB_PKGROOT_DIR}/include>")
	TARGET_INCLUDE_DIRECTORIES(${target} BEFORE PRIVATE "${NLIB_PKGROOT_DIR}/include" "${NLIB_PKGROOT_DIR}/sources")
	SET(win_no_def NOGDICAPMASKS NOVIRTUALKEYCODES NOWINMESSAGES NOWINSTYLES
		 NOSYSMETRICS NOMENUS NOICONS NOKEYSTATES NOSYSCOMMANDS NORASTEROPS NOSHOWWINDOW
		 OEMRESOURCE NOATOM NOCLIPBOARD NOCTRLMGR NODRAWTEXT NOGDI NOMB NOMETAFILE NOSCROLL
		 NOSERVICE NOSOUND NOTEXTMETRIC NOWH NOWINOFFSETS NOCOMM NOKANJI NOHELP
		 NODEFFERWINDOWPOS NOMCX)
	SET_PROPERTY(TARGET ${target} APPEND PROPERTY COMPILE_DEFINITIONS "${win_no_def}")
	IF(NOT ${CMAKE_GENERATOR} MATCHES "NMake Makefiles.*")
		SET_PROPERTY(TARGET ${target} APPEND PROPERTY COMPILE_OPTIONS "/MP")
	ENDIF()
	INSTALL(TARGETS ${target} DESTINATION lib/cmake/msvc)
	INSTALL(DIRECTORY ${NLIB_PKGROOT_DIR}/include DESTINATION . USE_SOURCE_PERMISSIONS)
ENDMACRO()

MACRO(target_link_libraries_default target)
	IF(NOT ${CMAKE_GENERATOR} MATCHES "NMake Makefiles.*")
		SET_PROPERTY(TARGET ${target} APPEND PROPERTY COMPILE_OPTIONS "/MP")
	ENDIF()
ENDMACRO()

GET_FILENAME_COMPONENT(
	NLIB_DIR
	"${NLIB_RELEASE_ROOT}/lib/cmake/msvc"
	ABSOLUTE)


