﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_BASE64_H_
#define INCLUDE_NN_NLIB_BASE64_H_

#include "nn/nlib/UniquePtr.h"

NLIB_NAMESPACE_BEGIN

class NLIB_VIS_PUBLIC Base64Encoder NLIB_FINAL {
 public:
    enum CharOption {
        kBase64PlusSlash = 0,
        kBase64PlusMinus,
        kBase64MinusUnderscore,
        kBase64DotMinus,
        kBase64UnderscoreColon,
        kBase64UnderscoreMinus,
        kBase64DotUnderscore,
        kBase64ExclamationMinus,
        kBase64Max,
        kBase64Default = kBase64PlusSlash,
        kBase64UrlSafe = kBase64MinusUnderscore,
        BASE64_PLUS_SLASH = kBase64PlusSlash,
        BASE64_PLUS_MINUS = kBase64PlusMinus,
        BASE64_MINUS_UNDERSCORE = kBase64MinusUnderscore,
        BASE64_DOT_MINUS = kBase64DotMinus,
        BASE64_UNDERSCORE_COLON = kBase64UnderscoreColon,
        BASE64_UNDERSCORE_MINUS = kBase64UnderscoreMinus,
        BASE64_DOT_UNDERSCORE = kBase64DotUnderscore,
        BASE64_EXCLAMATION_MINUS = kBase64ExclamationMinus,
        _BASE64_MAX = kBase64Max,
        BASE64_DEFAULT = kBase64PlusSlash,
        BASE64_URL_SAFE = kBase64MinusUnderscore
    };

 public:
    static size_t GetRequiredSize(size_t srcsize) NLIB_NOEXCEPT {
        return ((srcsize + 2) / 3) * 4 + 1;
    }
    static errno_t Encode(char* dst,
                          size_t dstsize,
                          const void* src,
                          size_t srcsize,
                          CharOption char_option = kBase64Default,
                          bool padding = false) NLIB_NOEXCEPT NLIB_NONNULL;
    template<size_t N>
    static errno_t Encode(char (&dst)[N],
                          const void* src,
                          size_t srcsize,
                          CharOption char_option = kBase64Default,
                          bool padding = false) NLIB_NOEXCEPT {
        return Encode(dst, N, src, srcsize, char_option, padding);
    }
#ifdef NLIB_CXX11_DEFAULT_TEMPLATE_ARGUMENT_FOR_FUNCTION_TEMPLATES
    template<class DUMMY = void>
#endif
    static errno_t Encode(UniquePtr<char[]>& dst,  // NOLINT
                          const void* src,
                          size_t srcsize,
                          CharOption char_option = kBase64Default,
                          bool padding = false) NLIB_NOEXCEPT {
        // to avoid crossing the DLL boundary
        size_t dstsize = GetRequiredSize(srcsize);
        // do not use UniquePtr to avoid C4714
        char* p = new (std::nothrow) char[dstsize];
        if (!p) return ENOMEM;
        errno_t e = Encode(p, dstsize, src, srcsize, char_option, padding);
        if (e == 0)
            dst.reset(p);
        else
            delete[] p;
        return e;
    }
    Base64Encoder() NLIB_NOEXCEPT : rem_(-1) {}
    ~Base64Encoder() NLIB_NOEXCEPT {}
    errno_t Init(CharOption char_option = kBase64Default) NLIB_NOEXCEPT;

    // note that dst is NOT null terminated
    errno_t StepEncode(
        size_t* written,
        char* dst,
        size_t dstsize,
        const void* src,
        size_t srcsize) NLIB_NOEXCEPT NLIB_NONNULL;
    // note that dst is NOT null terminated
    template<size_t N>
    errno_t StepEncode(
        size_t* written,
        char (&dst)[N],
        const void* src,
        size_t srcsize) NLIB_NOEXCEPT {
        return StepEncode(written, dst, N, src, srcsize);
    }
    // note that dst is NOT null terminated
    errno_t Close(
        size_t* written,
        char* dst,
        size_t dstsize,
        bool padding) NLIB_NOEXCEPT NLIB_NONNULL;
    // note that dst is NOT null terminated
    template<size_t N>
    errno_t Close(
        size_t* written,
        char (&dst)[N],
        bool padding) NLIB_NOEXCEPT {
        return Close(written, dst, N, padding);
    }
    NLIB_SAFE_BOOL(Base64Encoder, (rem_ >= 0))

 private:
    int rem_;
    uint8_t buf_[2];
    char table_[64];
};

class NLIB_VIS_PUBLIC Base64Decoder NLIB_FINAL {
 public:
    enum CharOption {
        kBase64PlusSlash = 0,
        kBase64PlusMinus,
        kBase64MinusUnderscore,
        kBase64DotMinus,
        kBase64UnderscoreColon,
        kBase64UnderscoreMinus,
        kBase64DotUnderscore,
        kBase64ExclamationMinus,
        kBase64Max,
        kBase64Default = kBase64PlusSlash,
        kBase64UrlSafe = kBase64MinusUnderscore,
        BASE64_PLUS_SLASH = kBase64PlusSlash,
        BASE64_PLUS_MINUS = kBase64PlusMinus,
        BASE64_MINUS_UNDERSCORE = kBase64MinusUnderscore,
        BASE64_DOT_MINUS = kBase64DotMinus,
        BASE64_UNDERSCORE_COLON = kBase64UnderscoreColon,
        BASE64_UNDERSCORE_MINUS = kBase64UnderscoreMinus,
        BASE64_DOT_UNDERSCORE = kBase64DotUnderscore,
        BASE64_EXCLAMATION_MINUS = kBase64ExclamationMinus,
        _BASE64_MAX = kBase64Max,
        BASE64_DEFAULT = kBase64PlusSlash,
        BASE64_URL_SAFE = kBase64MinusUnderscore
    };

 public:
    static size_t GetRequiredSize(size_t srcsize) NLIB_NOEXCEPT {
        return ((srcsize + 3) / 4) * 3;
    }
    static errno_t Decode(size_t* written,
                          uint8_t* dst,
                          size_t dstsize,
                          const char* src,
                          CharOption char_option = kBase64Default) NLIB_NOEXCEPT {
        return Decode_(written, dst, dstsize, src, nlib_strlen(src), char_option);
    }
    template<size_t N>
    static errno_t Decode(size_t* written,
                          uint8_t (&dst)[N],
                          const char* src,
                          CharOption char_option = kBase64Default) NLIB_NOEXCEPT {
        return Decode_(written, dst, N, src, nlib_strlen(src), char_option);
    }
#ifdef NLIB_CXX11_DEFAULT_TEMPLATE_ARGUMENT_FOR_FUNCTION_TEMPLATES
    template<class DUMMY = void>
#endif
    static errno_t Decode(UniquePtr<uint8_t[]>& dst,  // NOLINT
                          size_t* dstsize, const char* src,
                          CharOption char_option = kBase64Default) NLIB_NOEXCEPT {
        // to avoid crossing the DLL boundary
        size_t srcsize = nlib_strlen(src);
        size_t dstsize_ = GetRequiredSize(srcsize);
        // do not use UniquePtr to avoid C4714
        uint8_t* p = new (std::nothrow) uint8_t[dstsize_];
        if (!p) return ENOMEM;
        size_t written;
        errno_t e = Decode_(&written, p, dstsize_, src, srcsize, char_option);
        if (e == 0) {
            dst.reset(p);
            *dstsize = written;
        } else {
            delete[] p;
        }
        return e;
    }
    static errno_t DecodeInplace(
        size_t* written,
        char* src,
        CharOption char_option = kBase64Default) NLIB_NOEXCEPT NLIB_NONNULL;

    Base64Decoder() NLIB_NOEXCEPT : rem_(-1) {}
    ~Base64Decoder() NLIB_NOEXCEPT {}
    errno_t Init(CharOption char_option = kBase64Default) NLIB_NOEXCEPT;

    errno_t StepDecode(
        size_t* written,
        void* dst,
        size_t dstsize,
        const char* src,
        size_t srcsize) NLIB_NOEXCEPT NLIB_NONNULL;
    template<size_t N>
    errno_t StepDecode(
        size_t* written,
        uint8_t (&dst)[N],
        const char* src,
        size_t srcsize) NLIB_NOEXCEPT {
        return StepDecode(written, dst, N, src, srcsize);
    }
    errno_t Close(
        size_t* written,
        void* dst,
        size_t dstsize) NLIB_NOEXCEPT NLIB_NONNULL;
    template<size_t N>
    errno_t Close(
        size_t* written,
        uint8_t (&dst)[N]) NLIB_NOEXCEPT {
        return Close(written, dst, N);
    }
    NLIB_SAFE_BOOL(Base64Decoder, (rem_ >= 0))

 private:
    static errno_t Decode_(size_t* written,
                           uint8_t* dst,
                           size_t dstsize,
                           const char* src,
                           size_t srcsize,
                           CharOption char_option = kBase64Default) NLIB_NOEXCEPT NLIB_NONNULL;

    int rem_;
    int8_t buf_[4];
#if defined(NLIB_SIMD) && defined(NLIB_LITTLE_ENDIAN)
    char c62_;
    char c63_;
#endif
    int8_t table_[256];
};

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_BASE64_H_
