﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_LOCKFREE_H_
#define INCLUDE_NN_NLIB_LOCKFREE_H_

#include <new>
#include "nn/nlib/Config.h"
#include "nn/nlib/TypeTraits.h"
#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/DynamicAlignedStorage.h"
#include "nn/nlib/Swap.h"

NLIB_NAMESPACE_BEGIN

template<class T1, class T2 = void>
struct LockFreePlaceHolder {
    T1 data;  // T1 must be POD
    union next_type {
        uintptr_t aba_ptr;
        LockFreePlaceHolder<T1, T2>* raw_ptr;
    } next;
    T2 aux;
    void Init(const T1& elem) NLIB_NOEXCEPT NLIB_NO_TSAN {
        data = elem;
        aux = T2();
    }
};

template<class T1>
struct LockFreePlaceHolder<T1, void> {
    T1 data;  // T1 must be POD
    union next_type {
        uintptr_t aba_ptr;
        LockFreePlaceHolder<T1, void>* raw_ptr;
    } next;
    void Init(const T1& elem) NLIB_NOEXCEPT NLIB_NO_TSAN {
        data = elem;
    }
};

template<class T1, class T2 = void>
class LockFreePlaceHolderPool {
 public:
#if !defined(CAFE) && !defined(NN_PLATFORM_CTR)
    // some old compilers may complain that T1 is incomplete type even if it is not.
    NLIB_STATIC_ASSERT(IsPod<T1>::value);
#endif
    LockFreePlaceHolderPool() NLIB_NOEXCEPT : invariant_(0), mask_(0), mem_(NULL) {}
    ~LockFreePlaceHolderPool() NLIB_NOEXCEPT { delete[] mem_; }
    errno_t Init(size_t count) NLIB_NOEXCEPT;
    NLIB_ALWAYS_INLINE LockFreePlaceHolder<T1, T2>* CalcPtr(uintptr_t aba) const NLIB_NOEXCEPT {
        return reinterpret_cast<LockFreePlaceHolder<T1, T2>*>((aba & mask_) | invariant_);
    }
    NLIB_ALWAYS_INLINE bool IsNull(uintptr_t aba) const NLIB_NOEXCEPT {
        return aba & 1;
    }
    NLIB_ALWAYS_INLINE LockFreePlaceHolder<T1, T2>* Alloc() NLIB_NOEXCEPT NLIB_NO_TSAN {
        LockFreePlaceHolder<T1, T2>* ptr;
        uintptr_t expected = Load(&freelist_);
        for (;;) {
            if (NLIB_UNLIKELY(IsNull(expected))) return NULL;
            ptr = this->CalcPtr(expected);
            uintptr_t desired = Load(&ptr->next.aba_ptr);
            if (NLIB_LIKELY(Cas(&freelist_, &expected, &desired, 1)))
                break;
        }
        uintptr_t mask = mask_;
        ptr->next.aba_ptr = (ptr->next.aba_ptr & ~mask) + (mask + 2);
        return ptr;
    }
    NLIB_ALWAYS_INLINE void Free(LockFreePlaceHolder<T1, T2>* p) NLIB_NOEXCEPT NLIB_NO_TSAN {
        uintptr_t mask = mask_;
        uintptr_t desired = reinterpret_cast<uintptr_t>(p);
        uintptr_t expected = Load(&freelist_);
        uintptr_t base = (p->next.aba_ptr & ~mask) + (mask + 1);
        for (;;) {
            Store(&p->next.aba_ptr, base | (expected & mask));
            // must comfirm p->next written
            if (NLIB_LIKELY(Cas(&freelist_, &expected, &desired, 1, NLIB_ATOMIC_RELEASE)))
                break;
        }
    }
    NLIB_ALWAYS_INLINE LockFreePlaceHolder<T1, T2>* AllocUnsafe() NLIB_NOEXCEPT {
        uintptr_t expected = freelist_;
        if (NLIB_UNLIKELY(IsNull(expected))) return NULL;
        LockFreePlaceHolder<T1, T2>* ptr = this->CalcPtr(expected);
        freelist_ = ptr->next.aba_ptr;
        ptr->next.aba_ptr = 1;
        return ptr;
    }
    NLIB_ALWAYS_INLINE void FreeUnsafe(LockFreePlaceHolder<T1, T2>* p) NLIB_NOEXCEPT {
        p->next.aba_ptr = freelist_;
        freelist_ = reinterpret_cast<uintptr_t>(p);
    }
    NLIB_ALWAYS_INLINE uintptr_t Load(uintptr_t* ptr) NLIB_NOEXCEPT NLIB_NO_TSAN {
        void* val = nlib_atomic_loadptr(reinterpret_cast<void**>(ptr), NLIB_ATOMIC_RELAXED);
        return reinterpret_cast<uintptr_t>(val);
    }
    NLIB_ALWAYS_INLINE uintptr_t Load(uintptr_t* ptr, int memorder) NLIB_NOEXCEPT NLIB_NO_TSAN {
        void* val = nlib_atomic_loadptr(reinterpret_cast<void**>(ptr), memorder);
        return reinterpret_cast<uintptr_t>(val);
    }
    NLIB_ALWAYS_INLINE void Store(uintptr_t* ptr, uintptr_t value) NLIB_NOEXCEPT NLIB_NO_TSAN {
        nlib_atomic_storeptr(reinterpret_cast<void**>(ptr),
                             reinterpret_cast<void*>(value),
                             NLIB_ATOMIC_RELAXED);
    }
    NLIB_ALWAYS_INLINE void Store(uintptr_t* ptr, uintptr_t value, int memorder)
        NLIB_NOEXCEPT NLIB_NO_TSAN {
        nlib_atomic_storeptr(reinterpret_cast<void**>(ptr),
                             reinterpret_cast<void*>(value),
                             memorder);
    }
    NLIB_ALWAYS_INLINE int Cas(uintptr_t* ptr, uintptr_t* expected, uintptr_t* desired, int weak,
                               int success_memorder, int failure_memorder)
        NLIB_NOEXCEPT NLIB_NO_TSAN {
        uintptr_t mask = mask_;
        *desired = ((*desired & mask) | (mask + 1)) + (*expected & ~mask);
        int rval = nlib_atomic_compare_exchangeptr(reinterpret_cast<void**>(ptr),
                                                   reinterpret_cast<void**>(expected),
                                                   reinterpret_cast<void*>(*desired),
                                                   weak, success_memorder, failure_memorder);
        return rval;
    }
    NLIB_ALWAYS_INLINE int Cas(uintptr_t* ptr, uintptr_t* expected, uintptr_t* desired, int weak,
                               int success_memorder) NLIB_NOEXCEPT {
        return Cas(ptr, expected, desired, weak, success_memorder, NLIB_ATOMIC_RELAXED);
    }
    NLIB_ALWAYS_INLINE int Cas(uintptr_t* ptr, uintptr_t* expected, uintptr_t* desired,
                               int weak) NLIB_NOEXCEPT {
        return Cas(ptr, expected, desired, weak, NLIB_ATOMIC_RELAXED, NLIB_ATOMIC_RELAXED);
    }
    void SwapUnsafe(LockFreePlaceHolderPool& rhs) NLIB_NOEXCEPT {  // NOLINT
        std::swap(freelist_, rhs.freelist_);
        std::swap(invariant_, rhs.invariant_);
        std::swap(mask_, rhs.mask_);
        std::swap(mem_, rhs.mem_);
    }

 public:
    errno_t Push(uintptr_t* stack, const T1& elem) NLIB_NOEXCEPT;
    errno_t Pop(uintptr_t* stack, T1* elem) NLIB_NOEXCEPT;
    errno_t Enqueue(uintptr_t* tail, const T1& elem) NLIB_NOEXCEPT;
    errno_t Dequeue(uintptr_t* head, uintptr_t* tail, T1* elem) NLIB_NOEXCEPT;
    errno_t PushUnsafe(uintptr_t* stack, const T1& elem) NLIB_NOEXCEPT {
        typedef LockFreePlaceHolder<T1, T2> item_type;
        item_type* item_ptr = AllocUnsafe();
        if (NLIB_UNLIKELY(!item_ptr)) return EAGAIN;

        // point of no failure
        item_ptr->data = elem;
        item_ptr->next.aba_ptr = *stack;
        *stack = reinterpret_cast<uintptr_t>(item_ptr);
        return 0;
    }
    errno_t PopUnsafe(uintptr_t* stack, T1* elem) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(IsNull(*stack))) return EAGAIN;
        LockFreePlaceHolder<T1, T2>* ptr = CalcPtr(*stack);
        *elem = ptr->data;
        *stack = ptr->next.aba_ptr;
        FreeUnsafe(ptr);
        return 0;
    }
    errno_t EnqueueUnsafe(uintptr_t* tail, const T1& elem) NLIB_NOEXCEPT {
        LockFreePlaceHolder<T1, T2>* item_ptr = AllocUnsafe();
        if (NLIB_UNLIKELY(!item_ptr)) return EAGAIN;

        // point of no failure
        item_ptr->data = elem;
        CalcPtr(*tail)->next.aba_ptr = reinterpret_cast<uintptr_t>(item_ptr);
        *tail = reinterpret_cast<uintptr_t>(item_ptr);
        return 0;
    }
    errno_t DequeueUnsafe(uintptr_t* head, uintptr_t* tail, T1* elem) NLIB_NOEXCEPT {
        typedef LockFreePlaceHolder<T1, T2> item_type;
        item_type* first_ptr = CalcPtr(*head);
        item_type* last_ptr = CalcPtr(*tail);
        if (first_ptr == last_ptr) return EAGAIN;

        uintptr_t first_next = first_ptr->next.aba_ptr;
        item_type* first_next_ptr = CalcPtr(first_next);
        *elem = first_next_ptr->data;
        *head = first_next;
        FreeUnsafe(first_ptr);
        return 0;
    }

 private:
    uintptr_t freelist_;
    uintptr_t invariant_;
    uintptr_t mask_;
    LockFreePlaceHolder<T1, T2>* mem_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(LockFreePlaceHolderPool);
};

template<class T1, class T2>
errno_t LockFreePlaceHolderPool<T1, T2>::Init(size_t count) NLIB_NOEXCEPT {
    if (NLIB_UNLIKELY(mem_)) return EALREADY;
    LockFreePlaceHolder<T1, T2>* mem = new (std::nothrow) LockFreePlaceHolder<T1, T2>[count];
    if (NLIB_UNLIKELY(!mem)) return ENOMEM;
    uintptr_t freelist = 1;
#ifdef NLIB_64BIT
    for (size_t j = 0; j < 4; ++j) {
        for (size_t i = j; i < count; i += 4) {
            mem[i].next.aba_ptr = freelist;
            mem[i].data = T1();
            freelist = reinterpret_cast<uintptr_t>(&mem[i]);
        }
    }
#else
    for (size_t j = 0; j < 8; ++j) {
        for (size_t i = j; i < count; i += 8) {
            mem[i].next.aba_ptr = freelist;
            mem[i].data = T1();
            freelist = reinterpret_cast<uintptr_t>(&mem[i]);
        }
    }
#endif
    uintptr_t first = reinterpret_cast<uintptr_t>(mem);
    uintptr_t last = reinterpret_cast<uintptr_t>(mem + count);
#ifdef NLIB_64BIT
    int sft = nlib_clz64(first ^ last);
#else
    int sft = nlib_clz32(first ^ last);
#endif
    freelist_ = freelist;
    mask_ = (static_cast<uintptr_t>(0) - 1) >> sft;
    invariant_ = first & ~mask_;
    mem_ = mem;
    return 0;
}

template<class T1, class T2>
NLIB_ALWAYS_INLINE errno_t LockFreePlaceHolderPool<T1, T2>::Push(
    uintptr_t* stack, const T1& elem) NLIB_NOEXCEPT NLIB_NO_TSAN {
    typedef LockFreePlaceHolder<T1, T2> item_type;
    item_type* item_ptr = Alloc();
    if (NLIB_UNLIKELY(!item_ptr)) return EAGAIN;

    // point of no failure
    item_ptr->data = elem;
    uintptr_t mask = mask_;
    uintptr_t desired = reinterpret_cast<uintptr_t>(item_ptr);
    uintptr_t expected = Load(&freelist_);
    uintptr_t base = (item_ptr->next.aba_ptr & ~mask) + (mask + 1);
    for (;;) {
        Store(&item_ptr->next.aba_ptr, base | (expected & mask));
        // must comfirm p->next written
        if (NLIB_LIKELY(Cas(stack, &expected, &desired, 1, NLIB_ATOMIC_RELEASE)))
            break;
    }
    return 0;
}

template<class T1, class T2>
NLIB_ALWAYS_INLINE errno_t LockFreePlaceHolderPool<T1, T2>::Pop(
    uintptr_t* stack, T1* elem) NLIB_NOEXCEPT NLIB_NO_TSAN {
    LockFreePlaceHolder<T1, T2>* ptr;
    uintptr_t expected = Load(stack);
    for (;;) {
        if (NLIB_UNLIKELY(IsNull(expected))) return EAGAIN;
        ptr = this->CalcPtr(expected);
        uintptr_t desired = Load(&ptr->next.aba_ptr);
        if (NLIB_LIKELY(Cas(stack, &expected, &desired, 1)))
            break;
    }
    *elem = ptr->data;
    Free(ptr);
    return 0;
}

template<class T1, class T2>
NLIB_ALWAYS_INLINE errno_t LockFreePlaceHolderPool<T1, T2>::Enqueue(
    uintptr_t* tail, const T1& elem) NLIB_NOEXCEPT NLIB_NO_TSAN {
    LockFreePlaceHolder<T1, T2>* item_ptr = Alloc();
    if (NLIB_UNLIKELY(!item_ptr)) return EAGAIN;

    // point of no failure
    item_ptr->Init(elem);
#if !defined(_M_IX86) && !defined(_M_AMD64) && !defined(__i386__) && !defined(__x86_64__)
    nlib_atomic_thread_fence(NLIB_ATOMIC_RELEASE);
#endif
    uintptr_t last = Load(tail);
    for (;;) {
        uintptr_t* last_next_ptr = &CalcPtr(last)->next.aba_ptr;
        uintptr_t last_next = Load(last_next_ptr,
                                   NLIB_ATOMIC_ACQUIRE);  // before last_new is loaded
        uintptr_t last_new = Load(tail);
        if (NLIB_LIKELY(last == last_new)) {
            if (NLIB_LIKELY(IsNull(last_next))) {
                if (NLIB_LIKELY(Cas(last_next_ptr, &last_next,
                                    reinterpret_cast<uintptr_t*>(&item_ptr),
                                    1, NLIB_ATOMIC_ACQUIRE))) {
                    // Update tail
                    Cas(tail, &last, reinterpret_cast<uintptr_t*>(&item_ptr), 0);
                    return 0;
                }
            } else {
                // Update tail and retry
                if (NLIB_LIKELY(Cas(tail, &last, &last_next, 0))) {
                    last = last_next;
                }
            }
        } else {
            last = last_new;
        }
    }
}

template<class T1, class T2>
NLIB_ALWAYS_INLINE errno_t LockFreePlaceHolderPool<T1, T2>::Dequeue(
    uintptr_t* head, uintptr_t* tail, T1* elem) NLIB_NOEXCEPT NLIB_NO_TSAN {
    typedef LockFreePlaceHolder<T1, T2> item_type;
    // Load(head), then Load(tail)
    uintptr_t first = Load(head, NLIB_ATOMIC_ACQUIRE);
    uintptr_t last = Load(tail);
    for (;;) {
        item_type* first_ptr = CalcPtr(first);
        item_type* last_ptr = CalcPtr(last);
        uintptr_t first_next = Load(&first_ptr->next.aba_ptr,
                                    NLIB_ATOMIC_ACQUIRE);  // before first_new is loaded
        uintptr_t first_new = Load(head);
        if (NLIB_LIKELY(first == first_new)) {
            if (NLIB_UNLIKELY(first_ptr == last_ptr)) {
                if (NLIB_UNLIKELY(IsNull(first_next))) return EAGAIN;
                // Update tail and retry
                first = Load(head, NLIB_ATOMIC_ACQUIRE);
                if (NLIB_LIKELY(Cas(tail, &last, &first_next, 0))) {
                    last = first_next;
                }
            } else {
                item_type* first_next_ptr = CalcPtr(first_next);
                *elem = first_next_ptr->data;
                if (NLIB_LIKELY(Cas(head, &first, &first_next, 0))) {
                    Free(CalcPtr(first));
                    return 0;
                }
                first = Load(head, NLIB_ATOMIC_ACQUIRE);
                last = Load(tail);
            }
        } else {
            first = first_new;
            last = Load(tail);
        }
    }
}

template<size_t N>
class LockFreePipe {
 public:
    LockFreePipe() NLIB_NOEXCEPT : ofs_read_(0), ofs_write_(0) {}
    size_t GetBufferSize() NLIB_NOEXCEPT {
        NLIB_STATIC_ASSERT(N > 0);
        NLIB_STATIC_ASSERT((N & (N - 1)) == 0);
        NLIB_STATIC_ASSERT(N < INT32_MAX);
        return N;
    }
    NLIB_ALWAYS_INLINE errno_t Read(void* dest, size_t nbytes) NLIB_NOEXCEPT {
        uint32_t rofs = this->Load(&ofs_read_, NLIB_ATOMIC_RELAXED);
        uint32_t wofs = this->Load(&ofs_write_, NLIB_ATOMIC_ACQUIRE);
        // read ofs_write_, then read buffer_[rofs_actual]

        if (NLIB_UNLIKELY(nbytes > wofs - rofs)) return EAGAIN;
        uint32_t rofs_actual = rofs & (N - 1);

        size_t ntail = N - rofs_actual;
        if (ntail >= nbytes) {
            // ARMCC needs this cast.
            nlib_memcpy(dest,
                        nbytes,
                        reinterpret_cast<const void*>(&buffer_[rofs_actual]),
                        nbytes);
        } else {
            nlib_memcpy(dest,
                        nbytes,
                        reinterpret_cast<const void*>(&buffer_[rofs_actual]),
                        ntail);
            nlib_memcpy(reinterpret_cast<void*>(reinterpret_cast<uint8_t*>(dest) + ntail),
                        nbytes - ntail,
                        reinterpret_cast<void*>(&buffer_[0]),
                        nbytes - ntail);
        }
        rofs += static_cast<uint32_t>(nbytes);

        // must not be reordered
        this->Store(&ofs_read_, rofs, NLIB_ATOMIC_RELEASE);
        return 0;
    }
    NLIB_ALWAYS_INLINE errno_t Write(const void* src, size_t nbytes) NLIB_NOEXCEPT {
        uint32_t wofs = this->Load(&ofs_write_, NLIB_ATOMIC_RELAXED);
        uint32_t rofs = this->Load(&ofs_read_, NLIB_ATOMIC_ACQUIRE);
        // read ofs_read_, then write buffer_[wofs_actual]

        if (NLIB_UNLIKELY(nbytes > N - (wofs - rofs))) return EAGAIN;
        uint32_t wofs_actual = wofs & (N - 1);

        size_t ntail = N - wofs_actual;
        if (ntail >= nbytes) {
            nlib_memcpy(reinterpret_cast<void*>(&buffer_[wofs_actual]),
                        nbytes,
                        src,
                        nbytes);
        } else {
            nlib_memcpy(reinterpret_cast<void*>(&buffer_[wofs_actual]),
                        ntail,
                        src,
                        ntail);
            nlib_memcpy(reinterpret_cast<void*>(&buffer_[0]),
                        nbytes - ntail,
                        reinterpret_cast<const void*>(
                            reinterpret_cast<const uint8_t*>(src) + ntail),
                        nbytes - ntail);
        }
        wofs += static_cast<uint32_t>(nbytes);

        // must not be reordered
        this->Store(&ofs_write_, wofs, NLIB_ATOMIC_RELEASE);
        return 0;
    }

 private:
    NLIB_ALWAYS_INLINE uint32_t Load(uint32_t* p, int memorder) NLIB_NOEXCEPT {
        return static_cast<uint32_t>(nlib_atomic_load32(
            reinterpret_cast<int32_t*>(p), memorder));
    }
    NLIB_ALWAYS_INLINE void Store(uint32_t* p, uint32_t value, int memorder) NLIB_NOEXCEPT {
        nlib_atomic_store32(reinterpret_cast<int32_t*>(p),
            static_cast<int32_t>(value), memorder);
    }

    uint32_t ofs_read_;
    uint32_t ofs_write_;
    uint8_t buffer_[N];
    NLIB_DISALLOW_COPY_AND_ASSIGN(LockFreePipe);
};

template<class T>
class ConstructorForLockFree {
 public:
    T* operator()() NLIB_NOEXCEPT { return new (std::nothrow) T(); }
};

template<class T>
class DestructorForLockFree {
 public:
    void operator()(T* ptr) NLIB_NOEXCEPT { delete ptr; }
};

template<class T>
class LockFreeStack {
 public:
    typedef T* PopType;
    LockFreeStack() NLIB_NOEXCEPT : top_(0), pool_() {}
    ~LockFreeStack() NLIB_NOEXCEPT {
        // T is POD, and not pointer
    }
    // NOT thread safe
    errno_t Init(size_t count) NLIB_NOEXCEPT {
        errno_t e = pool_.Init(count);
        if (NLIB_UNLIKELY(e != 0)) return e;
        top_ = 1;
        return 0;
    }
    errno_t Push(const T& x) NLIB_NOEXCEPT { return pool_.Push(&top_, x); }
    errno_t Pop(PopType x) NLIB_NOEXCEPT { return pool_.Pop(&top_, x); }
    // NOT thread safe
    errno_t PushUnsafe(const T& x) NLIB_NOEXCEPT { return pool_.PushUnsafe(&top_, x); }
    // NOT thread safe
    errno_t PopUnsafe(PopType x) NLIB_NOEXCEPT { return pool_.PopUnsafe(&top_, x); }
    void SwapUnsafe(LockFreeStack& rhs) NLIB_NOEXCEPT {  // NOLINT
        std::swap(top_, rhs.top_);
        pool_.SwapUnsafe(rhs.pool_);
    }

 private:
    uintptr_t top_;
    LockFreePlaceHolderPool<T> pool_;
};

template<class T>
class LockFreeStack<T*> {
 public:
    typedef UniquePtr<T, DestructorForLockFree<T> > PopType;
    LockFreeStack() NLIB_NOEXCEPT : top_(0), pool_() {}
    ~LockFreeStack() NLIB_NOEXCEPT {
        if (top_ != 0) {
            T* elem;
            while (0 == pool_.PopUnsafe(&top_, &elem)) {
                DestructorForLockFree<T>()(elem);
            }
        }
    }
    // NOT thread safe
    errno_t Init(size_t count) NLIB_NOEXCEPT {
        errno_t e = pool_.Init(count);
        if (NLIB_UNLIKELY(e != 0)) return e;
        top_ = 1;
        return 0;
    }
    errno_t Push(T* obj) NLIB_NOEXCEPT { return pool_.Push(&top_, obj); }
    errno_t Pop(PopType& obj) NLIB_NOEXCEPT {  // NOLINT
        T* x;
        errno_t e = pool_.Pop(&top_, &x);
        if (e != 0) return e;
        obj.reset(x);
        return 0;
    }
    // NOT thread safe
    errno_t PushUnsafe(T* obj) NLIB_NOEXCEPT { return pool_.PushUnsafe(&top_, obj); }
    // NOT thread safe
    errno_t PopUnsafe(PopType& obj) NLIB_NOEXCEPT {  // NOLINT
        T* x;
        errno_t e = pool_.PopUnsafe(&top_, &x);
        if (e != 0) return e;
        obj.reset(x);
        return 0;
    }
    void SwapUnsafe(LockFreeStack& rhs) NLIB_NOEXCEPT {  // NOLINT
        std::swap(top_, rhs.top_);
        pool_.SwapUnsafe(rhs.pool_);
    }

 private:
    uintptr_t top_;
    LockFreePlaceHolderPool<T*> pool_;
};

template<class T>
class LockFreeQueue {
 public:
    typedef T* DequeueType;
    LockFreeQueue() NLIB_NOEXCEPT : head_(0), tail_(0), pool_() {}
    ~LockFreeQueue() NLIB_NOEXCEPT {
        // T is POD, and not pointer
    }
    // NOT thread safe
    errno_t Init(size_t count) NLIB_NOEXCEPT {
        errno_t e = pool_.Init(count + 1);
        if (NLIB_UNLIKELY(e != 0)) return e;
        head_ = tail_ = reinterpret_cast<uintptr_t>(pool_.AllocUnsafe());
        return 0;
    }
    errno_t Enqueue(const T& x) NLIB_NOEXCEPT { return pool_.Enqueue(&tail_, x); }
    errno_t Dequeue(DequeueType x) NLIB_NOEXCEPT { return pool_.Dequeue(&head_, &tail_, x); }
    // NOT thread safe
    errno_t EnqueueUnsafe(const T& x) NLIB_NOEXCEPT { return pool_.EnqueueUnsafe(&tail_, x); }
    // NOT thread safe
    errno_t DequeueUnsafe(DequeueType x) NLIB_NOEXCEPT {
        return pool_.DequeueUnsafe(&head_, &tail_, x);
    }
    void SwapUnsafe(LockFreeQueue& rhs) NLIB_NOEXCEPT {  // NOLINT
        std::swap(head_, rhs.head_);
        std::swap(tail_, rhs.tail_);
        pool_.SwapUnsafe(rhs.pool_);
    }

 private:
    uintptr_t head_;
    uintptr_t tail_;
    LockFreePlaceHolderPool<T> pool_;
};

template<class T>
class LockFreeQueue<T*> {
 public:
    typedef UniquePtr<T, DestructorForLockFree<T> > DequeueType;
    LockFreeQueue() NLIB_NOEXCEPT : head_(0), tail_(0), pool_() {}
    ~LockFreeQueue() NLIB_NOEXCEPT {
        if (head_ != 0) {
            T* elem;
            while (0 == pool_.DequeueUnsafe(&head_, &tail_, &elem)) {
                DestructorForLockFree<T>()(elem);
            }
        }
    }
    // NOT thread safe
    errno_t Init(size_t count) NLIB_NOEXCEPT {
        errno_t e = pool_.Init(count + 1);
        if (NLIB_UNLIKELY(e != 0)) return e;
        head_ = tail_ = reinterpret_cast<uintptr_t>(pool_.AllocUnsafe());
        return 0;
    }
    errno_t Enqueue(T* obj) NLIB_NOEXCEPT { return pool_.Enqueue(&tail_, obj); }
    errno_t Dequeue(DequeueType& obj) NLIB_NOEXCEPT {  // NOLINT
        T* x;
        errno_t e = pool_.Dequeue(&head_, &tail_, &x);
        if (e != 0) return e;
        obj.reset(x);
        return 0;
    }
    // NOT thread safe
    errno_t EnqueueUnsafe(T* x) NLIB_NOEXCEPT { return pool_.EnqueueUnsafe(&tail_, x); }
    // NOT thread safe
    errno_t DequeueUnsafe(DequeueType& obj) NLIB_NOEXCEPT {  // NOLINT
        T* x;
        errno_t e = pool_.DequeueUnsafe(&head_, &tail_, &x);
        if (e != 0) return e;
        obj.reset(x);
        return 0;
    }
    void SwapUnsafe(LockFreeQueue& rhs) NLIB_NOEXCEPT {  // NOLINT
        std::swap(head_, rhs.head_);
        std::swap(tail_, rhs.tail_);
        pool_.SwapUnsafe(rhs.pool_);
    }

 private:
    uintptr_t head_;
    uintptr_t tail_;
    LockFreePlaceHolderPool<T*> pool_;
};

template<class T>
errno_t LockFreeInit(T** ptr) NLIB_NOEXCEPT NLIB_NONNULL;

template<class T>
inline errno_t LockFreeInit(T** ptr) NLIB_NOEXCEPT {
    NLIB_EINVAL_IFNULL(ptr);
    void* pold = nlib_atomic_loadptr(reinterpret_cast<void**>(ptr), NLIB_ATOMIC_RELAXED);
    if (pold) return 0;
    T* obj = ConstructorForLockFree<T>()();
    if (!obj) {
        pold = nlib_atomic_loadptr(reinterpret_cast<void**>(ptr), NLIB_ATOMIC_RELAXED);
        return pold ? 0 : EAGAIN;
    }
    if (nlib_atomic_compare_exchangeptr(reinterpret_cast<void**>(ptr),
                                        &pold, reinterpret_cast<void*>(obj), 0,
                                        NLIB_ATOMIC_RELEASE, NLIB_ATOMIC_RELAXED)) {
        return 0;
    } else {
        DestructorForLockFree<T>()(obj);
        return 0;
    }
}

template<class T>
class LockFreePriorityQueue {
 public:
    typedef UniquePtr<T, DestructorForLockFree<T> > DequeueType;
    LockFreePriorityQueue() NLIB_NOEXCEPT : mq_() {
        mq_.raw_handle = 0;
    }
    ~LockFreePriorityQueue() NLIB_NOEXCEPT {
        if (mq_.raw_handle != 0) (void)nlib_mq_close(mq_);
    }
    // NOT thread safe
    errno_t Init(size_t max_size) NLIB_NOEXCEPT {
        // not thread safe
        if (mq_.raw_handle != 0) return EALREADY;
        if (max_size > INT32_MAX) return EINVAL;
        nlib_mq_attr attr;
        attr.flag = NLIB_MQ_LOCKFREE;
        attr.max_msg = static_cast<int32_t>(max_size);
        attr.destructor = LockFreePriorityQueue::dtor;
        return nlib_mq_open(&mq_, &attr);
    }
    // NOT thread safe
    errno_t Close() NLIB_NOEXCEPT {
        // not thread safe
        nlib_mq mq = mq_;
        mq_.raw_handle = 0;
        return nlib_mq_close(mq);
    }
    errno_t Enqueue(T* obj, int prio) NLIB_NOEXCEPT {
        // thread safe
        return nlib_mq_send(mq_, obj, prio);
    }
    errno_t Dequeue(DequeueType& obj, int* prio) NLIB_NOEXCEPT {  // NOLINT
        // thread safe
        T* x;
        errno_t e = nlib_mq_receive(mq_, reinterpret_cast<nlib_mq_msg*>(&x), prio);
        if (e != 0) return e;
        obj.reset(x);
        return 0;
    }
    void SwapUnsafe(LockFreePriorityQueue& rhs) NLIB_NOEXCEPT {  // NOLINT
        nlib_mq mq = rhs.mq_;
        rhs.mq_ = mq_;
        mq_ = mq;
    }

 private:
    static void dtor(nlib_mq_msg msg) {
        T* obj = reinterpret_cast<T*>(msg);
        DestructorForLockFree<T>()(obj);
    }

    nlib_mq mq_;
};

namespace detail {

class LockFreeBroadcastQueueImpl {
    typedef LockFreePlaceHolder<const void*, int32_t> item_type;

 public:
    LockFreeBroadcastQueueImpl() NLIB_NOEXCEPT
        : head_(1), tail_(1), listener_count_(0), pool_(), listeners_() {}
    NLIB_VIS_PUBLIC ~LockFreeBroadcastQueueImpl() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC errno_t Init(size_t max_size, size_t listeners) NLIB_NOEXCEPT;
    errno_t Enqueue(const void* msg) NLIB_NOEXCEPT {
        // thread safe
        return pool_.Enqueue(&tail_, msg);
    }
    errno_t Dequeue(int32_t listener_id, const void** msg) NLIB_NOEXCEPT {
        // thread safe if listener_id is different
        if (NLIB_UNLIKELY(listener_id < 0) ||
            NLIB_UNLIKELY(listener_id >= listener_count_)) return ERANGE;
        uintptr_t& pos = listeners_[listener_id];
        item_type* item = pool_.CalcPtr(pos);
        uintptr_t next = item->next.aba_ptr;
        if (pool_.IsNull(next)) return ENOENT;
        *msg = pool_.CalcPtr(next)->data;
        pos = next;
        nlib_atomic_add_fetch32(&item->aux, 1, NLIB_ATOMIC_RELEASE);
        return 0;
    }
    NLIB_VIS_PUBLIC const void* CheckAndRemove() NLIB_NOEXCEPT;
    errno_t DeleteUnsafe(const void** msg) NLIB_NOEXCEPT {
        if (pool_.IsNull(head_)) return ENOENT;
        item_type* item = pool_.CalcPtr(head_);
        const void* p = item->data;
        head_ = item->next.aba_ptr;
        *msg = p;
        return 0;
    }
    size_t GetListenerCount() const NLIB_NOEXCEPT { return listener_count_; }
    void SwapUnsafe(LockFreeBroadcastQueueImpl& rhs) NLIB_NOEXCEPT {  // NOLINT
        std::swap(head_, rhs.head_);
        std::swap(tail_, rhs.tail_);
        std::swap(listener_count_, rhs.listener_count_);
        pool_.SwapUnsafe(rhs.pool_);
        listeners_.swap(rhs.listeners_);
    }

 private:
    uintptr_t Load(uintptr_t* p, int memorder = NLIB_ATOMIC_RELAXED) NLIB_NOEXCEPT {
        return reinterpret_cast<uintptr_t>(nlib_atomic_loadptr(reinterpret_cast<void**>(p),
                                           memorder));
    }

 private:
    uintptr_t head_;
    uintptr_t tail_;
    int32_t listener_count_;
    LockFreePlaceHolderPool<const void*, int32_t> pool_;
    UniquePtr<uintptr_t[]> listeners_;
};

}  // namespace detail

template<class T>
class LockFreeBroadcastQueue {
 public:
    struct empty_func {
        void operator()(const T*) {}
    };
    typedef UniquePtr<const T, empty_func> DequeueType;
    LockFreeBroadcastQueue() NLIB_NOEXCEPT : impl_() {}
    ~LockFreeBroadcastQueue() NLIB_NOEXCEPT {
        const void* p;
        while (impl_.DeleteUnsafe(&p) == 0) {
            DestructorForLockFree<const T>()(reinterpret_cast<const T*>(p));
        }
    }
    // NOT thread safe
    errno_t Init(size_t max_size, size_t listeners) NLIB_NOEXCEPT {
        // not thread safe
        return impl_.Init(max_size, listeners);
    }
    errno_t Enqueue(const T* obj) NLIB_NOEXCEPT {
        // thread safe
        return impl_.Enqueue(obj);
    }
    errno_t Dequeue(int32_t listener_id, DequeueType& obj) NLIB_NOEXCEPT {  // NOLINT
        // thread safe if listener_id is different
        const void* x;
        errno_t e = impl_.Dequeue(listener_id, &x);
        if (e != 0) return e;
        obj.reset(reinterpret_cast<const T*>(x));
        const T* p = reinterpret_cast<const T*>(impl_.CheckAndRemove());
        if (p) {
            DestructorForLockFree<const T>()(p);
        }
        return 0;
    }
    size_t GetListenerCount() const NLIB_NOEXCEPT {
        // thread safe
        return impl_.GetListenerCount();
    }
    void SwapUnsafe(LockFreeBroadcastQueue& rhs) NLIB_NOEXCEPT {  // NOLINT
        impl_.SwapUnsafe(rhs);
    }

 private:
    detail::LockFreeBroadcastQueueImpl impl_;
};

class LockFreeUnitHeap NLIB_FINAL {
 public:
    class MemHolder {
     public:
        void* Get() NLIB_NOEXCEPT { return ptr_; }

     private:
        void* const ptr_;
        NLIB_DISALLOW_COPY_AND_ASSIGN(MemHolder);
    };
    LockFreeUnitHeap() NLIB_NOEXCEPT : mem_(), pool_() {}
    // NOT thread safe
    NLIB_VIS_PUBLIC errno_t Init(size_t unit_size, size_t align,
                                 size_t count) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC ~LockFreeUnitHeap() NLIB_NOEXCEPT;

    // MemHolder* x = heap.Alloc();
    // void* ptr = x->Get();  // 'ptr' is what you want.
    NLIB_ALWAYS_INLINE MemHolder* Alloc() NLIB_NOEXCEPT {
        return reinterpret_cast<MemHolder*>(pool_.Alloc());
    }

    NLIB_ALWAYS_INLINE void Free(MemHolder* p) NLIB_NOEXCEPT {
        pool_.Free(reinterpret_cast<LockFreePlaceHolder<void*>*>(p));
    }

    // NOT thread safe
    // MemHolder* x = heap.Alloc();
    // void* ptr = x->Get();  // 'ptr' is what you want.
    NLIB_ALWAYS_INLINE MemHolder* AllocUnsafe() NLIB_NOEXCEPT {
        return reinterpret_cast<MemHolder*>(pool_.AllocUnsafe());
    }
    // NOT thread safe
    NLIB_ALWAYS_INLINE void FreeUnsafe(MemHolder* p) NLIB_NOEXCEPT {
        pool_.FreeUnsafe(reinterpret_cast<LockFreePlaceHolder<void*>*>(p));
    }
    void SwapUnsafe(LockFreeUnitHeap& rhs) NLIB_NOEXCEPT {  // NOLINT
        mem_.swap(rhs.mem_);
        pool_.SwapUnsafe(rhs.pool_);
    }

 private:
    DynamicAlignedStorage mem_;
    LockFreePlaceHolderPool<void*> pool_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(LockFreeUnitHeap);
};

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_LOCKFREE_H_
