﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_OUTPUTSTREAM_H_
#define INCLUDE_NN_NLIB_OUTPUTSTREAM_H_

#include "nn/nlib/Config.h"

NLIB_NAMESPACE_BEGIN

// code snippets:
//   OutputStream& stream = 'stream that inherits OutputStream';
//   if (!stream.Write(dataptr, nbyte)) { error = stream.GetErrorValue(); .... }
//   for (.....) { if (!stream.Write(byte)) { error = stream.GetErrorValue(); .... } }
//   # you should Flush() and Close() explicitly
//   if (!stream.Flush()) { error = stream.GetErrorValue(); .... } }
//   if (!stream.Close()) { error = stream.GetErrorValue(); .... } }
class NLIB_VIS_PUBLIC OutputStream {
 public:
    enum BufferingMode {
        kBufferingModeBlockBuffered = 0,  // default
        kBufferingModeLineBuffered,       // if console output
        kBufferingModeUnbuffered,
        BUFFERINGMODE_BLOCKBUFFERED = kBufferingModeBlockBuffered,
        BUFFERINGMODE_LINEBUFFERED = kBufferingModeLineBuffered,
        BUFFERINGMODE_UNBUFFERED = kBufferingModeUnbuffered
    };

 public:
    NLIB_CEXPR OutputStream() NLIB_NOEXCEPT : pos_(0),
                                              os_buf_curidx_(0),
                                              os_buf_(NULL),
                                              os_bufsize_(0),
                                              errno_(0),
                                              m_BufferingMode(kBufferingModeBlockBuffered) {}
    virtual ~OutputStream() NLIB_NOEXCEPT {
        // NOTE:
        // cannot call Close_() from this destructor.
        // the destructor of the derived class has to call Close_() if needed.
    }
    // use Pos64() if you need a 64bit value on 32bit arch
    size_t Pos() const NLIB_NOEXCEPT { return static_cast<size_t>(Pos64()); }
    uint64_t Pos64() const NLIB_NOEXCEPT { return pos_ + os_buf_curidx_; }
    bool Write(int b) NLIB_NOEXCEPT {
        // NOTE:
        // if the stream is closed,
        // Write() will succeed before the internal buffer becomes full.
        if (NLIB_UNLIKELY(os_buf_curidx_ == os_bufsize_)) {
            if (NLIB_UNLIKELY(!this->Flush_(false))) return false;
            if (NLIB_UNLIKELY(os_bufsize_ == 0)) {
                this->SetError(EIO);
                return false;
            }
        }
        os_buf_[os_buf_curidx_++] = static_cast<unsigned char>(b & 0xff);
        return true;
    }
    bool Write(const void* p, size_t n) NLIB_NOEXCEPT NLIB_NONNULL {
#ifndef NLIB_NONNULL_ENABLED
        if (!p) {
            this->SetError(EINVAL);
            return false;
        }
#endif
        if (n > RSIZE_MAX) {
            this->SetError(EINVAL);
            return false;
        }
        if (os_buf_curidx_ + n <= os_bufsize_) {
            nlib_memcpy(&os_buf_[os_buf_curidx_], n, p, n);
            os_buf_curidx_ += n;
            return true;
        }
        return this->Write_(p, n);
    }
    bool WriteGather(const nlib_fd_iovec* iov, int iovcnt) NLIB_NOEXCEPT NLIB_NONNULL {
#ifndef NLIB_NONNULL_ENABLED
        if (!iov) {
            this->SetError(EINVAL);
            return false;
        }
#endif
        if (iovcnt < 0) {
            this->SetError(EINVAL);
            return false;
        }
        return this->WriteGather_(iov, iovcnt);
    }
    bool Flush() NLIB_NOEXCEPT { return Flush_(true); }
    bool Close() NLIB_NOEXCEPT;
    errno_t GetErrorValue() const NLIB_NOEXCEPT { return errno_; }
    // returns true if there has been no errors
    // you can also write the codes such that 'if (!!stream) { ... }'
    BufferingMode GetBufferingMode() const NLIB_NOEXCEPT { return m_BufferingMode; }
    NLIB_SAFE_BOOL(OutputStream, errno_ == 0)

 protected:
    void ResetBuffer(void* p, size_t nbytes) NLIB_NOEXCEPT {
        os_buf_ = reinterpret_cast<unsigned char*>(p);
        os_bufsize_ = static_cast<int>(nbytes);
    }
    void SetError(errno_t e) const NLIB_NOEXCEPT {
        if (errno_ == 0) errno_ = e;
    }

 private:
    virtual bool PushBuffer_(const void* p, size_t nbytes, bool do_flush) NLIB_NOEXCEPT = 0;
    virtual bool Close_() NLIB_NOEXCEPT = 0;
    virtual void* GetWorkBuffer_(size_t* nbytes) NLIB_NOEXCEPT;
    virtual bool WriteGather_(const nlib_fd_iovec* iov, int iovcnt) NLIB_NOEXCEPT;
    bool Write_(const void* p, size_t n) NLIB_NOEXCEPT;

 private:
    bool Flush_(bool flush_device) NLIB_NOEXCEPT;
    bool GetWorkBuffer_() NLIB_NOEXCEPT;

 private:
    uint64_t pos_;
    size_t os_buf_curidx_;
    unsigned char* os_buf_;
    size_t os_bufsize_;
    mutable ErrnoT errno_;

 protected:
    // NOTE:
    // BufferingMode is for the codes which use OutputStream.
    // OutputStream only shows the buffering mode which the user specified.
    // for example, TextWriter (the user of OutputStream) has to decide
    // when to call Flush() looking at GetBufferingMode().
    BufferingMode m_BufferingMode;

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(OutputStream);
};

class NLIB_VIS_PUBLIC NullOutputStream NLIB_FINAL : public OutputStream {
 public:
    NullOutputStream() NLIB_NOEXCEPT;
    virtual ~NullOutputStream() NLIB_NOEXCEPT NLIB_OVERRIDE {}

 private:
    unsigned char dummy_buf_[256];
    virtual bool PushBuffer_(const void* p, size_t, bool do_flush) NLIB_NOEXCEPT NLIB_OVERRIDE;
    virtual bool Close_() NLIB_NOEXCEPT NLIB_OVERRIDE { return true; }
};

namespace detail {

template<size_t N>
class MiniBufOut {
 public:
    static NLIB_CEXPR const size_t buf_size = N;
    NLIB_ALWAYS_INLINE MiniBufOut() NLIB_NOEXCEPT { cur_ = &buf_[0]; }
    template<class OS>
    NLIB_ALWAYS_INLINE bool Flush(OS* stream) NLIB_NOEXCEPT {
        size_t n = cur_ - &buf_[0];
        cur_ = &buf_[0];
        return stream->Write(&buf_[0], n);
    }
    template<class OS>
    NLIB_ALWAYS_INLINE bool CheckAndFlush(OS* stream, size_t n) NLIB_NOEXCEPT {
        if (cur_ + n > &buf_[0] + N) {
            NLIB_ASSERT(n <= N);
            return Flush(stream);
        }
        return true;
    }
    NLIB_ALWAYS_INLINE void Advance(size_t n) NLIB_NOEXCEPT {
        // CheckAndFlush(stream, n) must be successful beforehand.
        cur_ += n;
    }
    NLIB_ALWAYS_INLINE uint8_t& operator[](size_t n) NLIB_NOEXCEPT {
        // CheckAndFlush(stream, n + 1) must be successful beforehand.
        return cur_[n];
    }
    template<class OS>
    NLIB_ALWAYS_INLINE bool WriteAndAdvance(OS* stream, uint8_t data) NLIB_NOEXCEPT {
        if (NLIB_LIKELY(CheckAndFlush(stream, 1))) {
            *cur_ = data;
            ++cur_;
            return true;
        }
        return false;
    }
    template<class OS>
    NLIB_ALWAYS_INLINE bool WriteAndAdvance(OS* stream,
                                            const void* data, size_t n) NLIB_NOEXCEPT {
        if (NLIB_LIKELY(CheckAndFlush(stream, n))) {
            (void)nlib_memcpy(cur_, n, data, n);
            cur_ += n;
            return true;
        }
        return false;
    }

 private:
    uint8_t* cur_;
    uint8_t buf_[N];
    NLIB_DISALLOW_COPY_AND_ASSIGN(MiniBufOut);
};

template<class T>
class MiniReallocOut {
 public:
    NLIB_ALWAYS_INLINE MiniReallocOut(T** mem, uint32_t* cur, uint32_t* memsize) NLIB_NOEXCEPT
        : mem_(mem), cur_(cur), memsize_(memsize) {}
    NLIB_ALWAYS_INLINE ~MiniReallocOut() NLIB_NOEXCEPT {}
    NLIB_ALWAYS_INLINE bool Resize(uint32_t n) NLIB_NOEXCEPT {
        if (n <= *memsize_) {
            *cur_ = n;
        } else {
            void* p = nlib_realloc(*mem_, n * sizeof(**mem_));
            if (!p) return false;
            *mem_ = reinterpret_cast<T*>(p);
            *memsize_ = n;
            *cur_ = n;
        }
        return true;
    }
    NLIB_ALWAYS_INLINE bool Prepare(uint32_t n) NLIB_NOEXCEPT {
        // T must be int, uint32_t, ... etc. for example
        if (NLIB_UNLIKELY(*cur_ + n > *memsize_)) {
            uint32_t newsize;
            const uint32_t t = 4096 / sizeof(**mem_);
            if (*cur_ + n >= t) {
                newsize = ((*cur_ + n) + t - 1) & ~(t - 1);
            } else {
                newsize = 1 << (32 - nlib_clz32(*cur_ + n - 1));
            }
            void* p = nlib_realloc(*mem_, newsize * sizeof(**mem_));
            if (!p) return false;
            *mem_ = reinterpret_cast<T*>(p);
            *memsize_ = newsize;
        }
        return true;
    }
    NLIB_ALWAYS_INLINE T& operator[](uint32_t n) NLIB_NOEXCEPT {
        // Prepare(n + 1) must be successful beforehand.
        return (*mem_)[*cur_ + n];
    }
    NLIB_ALWAYS_INLINE void Append(const T& v) NLIB_NOEXCEPT {
        // Prepare(1) must be successful beforehand.
        (*mem_)[*cur_] = v;
        ++(*cur_);
    }
    NLIB_ALWAYS_INLINE void Advance(uint32_t n) NLIB_NOEXCEPT {
        // Prepare(n) must be successful beforehand.
        *cur_ += n;
    }

 private:
    T** mem_;
    uint32_t* cur_;
    uint32_t* memsize_;  // sizeof(*mem) * memsize_ bytes
    NLIB_DISALLOW_COPY_AND_ASSIGN(MiniReallocOut);
};

}  // namespace detail

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_OUTPUTSTREAM_H_
