﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_SIMD_SIMDALGORITHM_H_
#define INCLUDE_NN_NLIB_SIMD_SIMDALGORITHM_H_

#include <algorithm>
#include <functional>
#include "nn/nlib/simd/SimdInt.h"
#include "nn/nlib/DynamicAlignedStorage.h"

#if defined(NLIB_SIMD)

NLIB_NAMESPACE_BEGIN
namespace simd {
namespace detail {

NLIB_VIS_PUBLIC void sortUint32A16(uint32_t* buf, uint32_t* src, size_t N) NLIB_NOEXCEPT;
NLIB_VIS_PUBLIC void sortUint32A16_1(uint32_t* buf, uint32_t* src, size_t N) NLIB_NOEXCEPT;

template <size_t NumElem, bool flag>
struct MergeSortHelper {
    static NLIB_ALWAYS_INLINE void Sort(uint32_t* data) NLIB_NOEXCEPT {
        NLIB_ALIGNAS(16) uint32_t tmp[NumElem];
        sortUint32A16(tmp, data, NumElem);
    }
};

template <size_t NumElem>
struct MergeSortHelper<NumElem, false> {
    static NLIB_ALWAYS_INLINE void Sort(uint32_t* data) NLIB_NOEXCEPT {
        NLIB_ALIGNAS(16) uint32_t tmp[NumElem];
        sortUint32A16_1(tmp, data, NumElem);
    }
};

}  // namespace detail

template <size_t NumElem>
// data must be 16 bytes aligned, NumElem must be multiple of 16
NLIB_ALWAYS_INLINE void MergeSortUint32A16(uint32_t* data) NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT((NumElem > 0) && (NumElem % 16 == 0));
    NLIB_ASSERT(!(reinterpret_cast<uintptr_t>(data) & 15));
    detail::MergeSortHelper<NumElem, ((NumElem & (NumElem - 1)) == 0)>::Sort(data);
}

NLIB_VIS_PUBLIC errno_t MergeSortUint32A16(uint32_t* data, size_t n) NLIB_NOEXCEPT;

template <class PRED>
// i128 PRED(i128 c) { mask(c.u8[i]) if matched }
const void* nlib_memchr_pred(const void* s, PRED pred, size_t n) NLIB_NOEXCEPT {
    if (!s) return NULL;

    const unsigned char* p = reinterpret_cast<const unsigned char*>(s);
    i128 a1, a2;
    i128 cmp1, cmp2;
    uint32_t mask;
    if (reinterpret_cast<uintptr_t>(p) & 15) {
        size_t r = reinterpret_cast<uintptr_t>(p) & 15;
        a1 = I128::LoadA16(p - r);
        cmp1 = pred(a1);
        mask = I128::MoveMask8(cmp1);
        mask = mask >> r;
        size_t rr = 16 - r;
        if (n < rr) {
            mask &= (1 << n) - 1;
            if (mask) return p + nlib_ctz32(mask);
            return NULL;
        }
        if (mask) return p + nlib_ctz32(mask);
        p += rr;
        n -= rr;
    }
    if (n >= 16) {
        if ((reinterpret_cast<uintptr_t>(p) & 32)) {
            a1 = I128::LoadA16(p);
            cmp1 = pred(a1);
            if (!I128::IsZero(cmp1)) {
                mask = I128::MoveMask8(cmp1);
                return p + nlib_ctz32(mask);
            }
            p += 16;
            n -= 16;
        }
        while (n >= 32) {
            a1 = I128::LoadA16(p);
            a2 = I128::LoadA16(p + 16);
            cmp1 = I128::SetZero();
            cmp2 = I128::SetZero();
            cmp1 = pred(a1);
            cmp2 = pred(a2);
            if (!I128::IsZero(I128::Or(cmp1, cmp2))) {
                mask = I128::MoveMask8(cmp1) | (I128::MoveMask8(cmp2) << 16);
                return p + nlib_ctz32(mask);
            }
            p += 32;
            n -= 32;
        }
        if (n >= 16) {
            a1 = I128::LoadA16(p);
            cmp1 = pred(a1);
            if (!I128::IsZero(cmp1)) {
                mask = I128::MoveMask8(cmp1);
                return p + nlib_ctz32(mask);
            }
            p += 16;
            n -= 16;
        }
    }
    if (n > 0) {
        a1 = I128::LoadA16(p);
        cmp1 = pred(a1);
        mask = I128::MoveMask8(cmp1);
        mask &= (1 << n) - 1;
        if (mask) return p + nlib_ctz32(mask);
    }
    return NULL;
}

template <class PRED>
// i128 PRED(i128 c) { mask(c.u8[i]) if matched }
const void* nlib_memchr_pred_not(const void* s, PRED pred, size_t n) NLIB_NOEXCEPT {
    if (!s) return NULL;

    const unsigned char* p = reinterpret_cast<const unsigned char*>(s);
    i128 a1, a2;
    i128 cmp1, cmp2;
    uint32_t mask;
    if (reinterpret_cast<uintptr_t>(p) & 15) {
        size_t r = reinterpret_cast<uintptr_t>(p) & 15;
        a1 = I128::LoadA16(p - r);
        cmp1 = I128::Not(pred(a1));
        mask = I128::MoveMask8(cmp1);
        mask = mask >> r;
        size_t rr = 16 - r;
        if (n < rr) {
            mask &= (1 << n) - 1;
            if (mask) return p + nlib_ctz32(mask);
            return NULL;
        }
        if (mask) return p + nlib_ctz32(mask);
        p += rr;
        n -= rr;
    }
    if (n >= 16) {
        if ((reinterpret_cast<uintptr_t>(p) & 32)) {
            a1 = I128::LoadA16(p);
            cmp1 = pred(a1);
            if (!I128::IsFull(cmp1)) {
                mask = I128::MoveMask8(I128::Not(cmp1));
                return p + nlib_ctz32(mask);
            }
            p += 16;
            n -= 16;
        }
        while (n >= 32) {
            a1 = I128::LoadA16(p);
            a2 = I128::LoadA16(p + 16);
            cmp1 = pred(a1);
            cmp2 = pred(a2);
            if (!I128::IsFull(I128::And(cmp1, cmp2))) {
                mask = I128::MoveMask8(I128::Not(cmp1)) | (I128::MoveMask8(I128::Not(cmp2)) << 16);
                return p + nlib_ctz32(mask);
            }
            p += 32;
            n -= 32;
        }
        if (n >= 16) {
            a1 = I128::LoadA16(p);
            cmp1 = pred(a1);
            if (!I128::IsFull(cmp1)) {
                mask = I128::MoveMask8(I128::Not(cmp1));
                return p + nlib_ctz32(mask);
            }
            p += 16;
            n -= 16;
        }
    }
    if (n > 0) {
        a1 = I128::LoadA16(p);
        cmp1 = I128::Not(pred(a1));
        mask = I128::MoveMask8(cmp1);
        mask &= (1 << n) - 1;
        if (mask) return p + nlib_ctz32(mask);
    }
    return NULL;
}

// mask if c.u8[i] is a-zA-Z
inline i128 __vectorcall IsAlpha(i128 c) NLIB_NOEXCEPT {
    i128 result = I128::CmpLtInt8(c, I128::SetValue('{', each_int8));
    result = I128::And(result, I128::CmpGtInt8(c, I128::SetValue('`', each_int8)));

    i128 tmp = I128::CmpLtInt8(c, I128::SetValue('[', each_int8));
    tmp = I128::And(tmp, I128::CmpGtInt8(c, I128::SetValue('@', each_int8)));
    result = I128::Or(result, tmp);

    return result;
}

// mask if c.u8[i] is 0-9
inline i128 __vectorcall IsDigit(i128 c) NLIB_NOEXCEPT {
    i128 result = I128::CmpLtInt8(c, I128::SetValue(':', each_int8));
    result = I128::And(result, I128::CmpGtInt8(c, I128::SetValue('/', each_int8)));

    return result;
}

// mask if c.u8[i] is a-zA-Z0-9
inline i128 __vectorcall IsAlnum(i128 c) NLIB_NOEXCEPT { return I128::Or(IsDigit(c), IsAlpha(c)); }

// mask if c.u8[i] is space, CR, LF, or tab
inline i128 __vectorcall IsSpace(i128 c) NLIB_NOEXCEPT {
    i128 result = I128::CmpEq8(c, I128::SetValue(' ', each_int8));
    result = I128::Or(result, I128::CmpEq8(c, I128::SetValue('\r', each_int8)));
    result = I128::Or(result, I128::CmpEq8(c, I128::SetValue('\n', each_int8)));
    result = I128::Or(result, I128::CmpEq8(c, I128::SetValue('\t', each_int8)));

    return result;
}

// mask if c.u8[i] is 0-9A-Fa-f
inline i128 __vectorcall IsXdigit(i128 c) NLIB_NOEXCEPT {
    i128 tmp;
    i128 result = I128::CmpLtInt8(c, I128::SetValue(':', each_int8));
    result = I128::And(result, I128::CmpGtInt8(c, I128::SetValue('/', each_int8)));

    tmp = I128::CmpLtInt8(c, I128::SetValue('G', each_int8));
    tmp = I128::And(tmp, I128::CmpGtInt8(c, I128::SetValue('@', each_int8)));
    result = I128::Or(result, tmp);

    tmp = I128::CmpLtInt8(c, I128::SetValue('g', each_int8));
    tmp = I128::And(tmp, I128::CmpGtInt8(c, I128::SetValue('`', each_int8)));
    result = I128::Or(result, tmp);

    return result;
}

namespace detail {
template<class T, class Compare>
class KeyIdxSortLess {
 public:
    KeyIdxSortLess(T* const* src, uint32_t mask,
                   Compare comp) NLIB_NOEXCEPT : src_(src), mask_(mask), comp_(comp) {
    }
    NLIB_ALWAYS_INLINE bool operator()(uint32_t lhs, uint32_t rhs) const NLIB_NOEXCEPT {
        return comp_(*src_[lhs & mask_], *src_[rhs & mask_]);
    }

 private:
    T* const* src_;
    uint32_t mask_;
    Compare comp_;
};

NLIB_ALIGNAS(16) extern NLIB_VIS_PUBLIC const uint32_t keyidxsort_0123[4];
}  // namespace detail

template<class T, class Compare>
errno_t KeyIdxSortN(T** dst, T* const* src, size_t n, Compare comp) NLIB_NOEXCEPT {
    // T must have a member function such that
    // uint32_t GetKey32() const;
#ifdef NLIB_64BIT
    if (n == 0 || n > 0x7FFFFFFFU) return EINVAL;
#else
    if (n == 0 || n > RSIZE_MAX) return EINVAL;
#endif
    errno_t e;
    size_t n16 = (n + 15) & ~15;

    DynamicAlignedStorage mem_;
    e = mem_.Init(n16 * sizeof(uint32_t), 16);  // NOLINT
    if (NLIB_UNLIKELY(e != 0)) return e;
    uint32_t* mem = reinterpret_cast<uint32_t*>(mem_.Get());
    int idx_width = 32 - nlib_clz32(static_cast<uint32_t>(n16));
    uint32_t idx_mask = (1U << idx_width) - 1;
    uint32_t idx_mask_inv = ~idx_mask;

    uint32_t max_key = 0U;
    uint32_t min_key = 0xFFFFFFFFU;
    for (size_t i = 0; i < n; ++i) {
        uint32_t key = src[i]->GetKey32();
        mem[i] = key;
        if (min_key > key)
            min_key = key;
        else if (max_key < key)
            max_key = key;
    }
    // several number of the bits could be omitted because they are the same.
    int left_shift = nlib_clz32(min_key ^ max_key);

    // embed indices
#if 1
    i128 vecmask = I128::SetValue(idx_mask_inv, each_uint32);
    i128 vecidx0 = I128::LoadA16(&detail::keyidxsort_0123[0]);
    i128 vec_i = I128::SetZero();
    i128 four = I128::SetValue(4, each_uint32);
    i128 vecidx4 = I128::Add32(vecidx0, four);
    i128 vecidx8 = I128::Add32(vecidx4, four);
    i128 vecidx12 = I128::Add32(vecidx8, four);
    i128 d16 = I128::Mult32(four, four);

    for (size_t i = 0; i < n16; i += 16) {
        i128 m0 = I128::LoadA16(&mem[i]);
        i128 m1 = I128::LoadA16(&mem[i + 4]);
        i128 m2 = I128::LoadA16(&mem[i + 8]);
        i128 m3 = I128::LoadA16(&mem[i + 12]);

        m0 = I128::And(I128::ShiftLeftLogical32(m0, left_shift), vecmask);
        m1 = I128::And(I128::ShiftLeftLogical32(m1, left_shift), vecmask);
        m2 = I128::And(I128::ShiftLeftLogical32(m2, left_shift), vecmask);
        m3 = I128::And(I128::ShiftLeftLogical32(m3, left_shift), vecmask);
        m0 = I128::Or(m0, I128::Add32(vec_i, vecidx0));
        m1 = I128::Or(m1, I128::Add32(vec_i, vecidx4));
        m2 = I128::Or(m2, I128::Add32(vec_i, vecidx8));
        m3 = I128::Or(m3, I128::Add32(vec_i, vecidx12));

        I128::StoreA16(&mem[i], m0);
        I128::StoreA16(&mem[i + 4], m1);
        I128::StoreA16(&mem[i + 8], m2);
        I128::StoreA16(&mem[i + 12], m3);
        vec_i = I128::Add16(vec_i, d16);
    }
#else
    for (size_t i = 0; i < n; ++i) {
        mem[i] = ((mem[i] << left_shift) & idx_mask_inv) | static_cast<uint32_t>(i);
    }
#endif
    for (size_t i = n; i < n16; ++i) {
        // It is ok because MergeSortUint32A16 is a stable sort algorithm.
        mem[i] = idx_mask_inv | static_cast<uint32_t>(i);
    }

    e = MergeSortUint32A16(mem, n16);
    if (NLIB_UNLIKELY(e != 0)) return e;

    // they must be sorted if the keys are the same.
    uint32_t prev_key = mem[0] & idx_mask_inv;
    size_t i = 1;
    detail::KeyIdxSortLess<T, Compare> myless(src, idx_mask, comp);
    while (i < n) {
        uint32_t key = mem[i] & idx_mask_inv;
        if (NLIB_UNLIKELY(key == prev_key)) {
            size_t from = i - 1;
            do {
                ++i;
                if (i == n) break;
                key = mem[i] & idx_mask_inv;
            } while (key == prev_key);
            // if the sort algorithm is not stable, mem[i -> n16] might be swapped.
            std::stable_sort(&mem[from], &mem[i], myless);
        }
        prev_key = key;
        ++i;
    }

    for (i = 0; i < n; ++i) {
        dst[i] = src[mem[i] & idx_mask];
    }
    return 0;
}

template<class T>
NLIB_ALWAYS_INLINE errno_t KeyIdxSortN(T** dst, T* const* src, size_t n) NLIB_NOEXCEPT {
    return KeyIdxSortN(dst, src, n, std::less<T>());
}

template<class T, class Compare>
inline errno_t KeyIdxSort(T** first, T** last, Compare comp) NLIB_NOEXCEPT {
    size_t n = last - first;
    T** tmp = reinterpret_cast<T**>(nlib_malloc(n * sizeof(*first)));
    if (NLIB_UNLIKELY(!tmp)) return ENOMEM;
    errno_t e = KeyIdxSortN(tmp, first, n, comp);
    if (NLIB_LIKELY(e == 0)) {
        nlib_memcpy(first, n * sizeof(*first), tmp, n * sizeof(*tmp));
    }
    nlib_free(tmp);
    return e;
}

template<class T>
NLIB_ALWAYS_INLINE errno_t KeyIdxSort(T** first, T** last) NLIB_NOEXCEPT {
    return KeyIdxSort(first, last, std::less<T>());
}

}  // namespace simd
NLIB_NAMESPACE_END

#endif

#endif  // INCLUDE_NN_NLIB_SIMD_SIMDALGORITHM_H_
