﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_THREADING_LIMITEDSHAREDPTR_H_
#define INCLUDE_NN_NLIB_THREADING_LIMITEDSHAREDPTR_H_

#include <new>

#include "nn/nlib/Swap.h"
#include "nn/nlib/UniquePtr.h"

NLIB_NAMESPACE_BEGIN
namespace threading {

namespace internal {

class LspStatus {
 public:
    NLIB_CEXPR LspStatus() NLIB_NOEXCEPT : value_(NULL) {}
    ~LspStatus() NLIB_NOEXCEPT {}
    NLIB_MOVE_MEMBER_HELPER_1(LspStatus, value_)
    errno_t ToShare(LspStatus* ptr) NLIB_NOEXCEPT {
        if (!value_) {
            value_ = new (std::nothrow) int32_t(2);
            if (!value_) return ENOMEM;
        } else {
            nlib_atomic_add_fetch32(value_, 1, NLIB_ATOMIC_ACQUIRE);
        }
        NLIB_ASSERT(!ptr->value_);
        ptr->value_ = value_;
        return 0;
    }
    bool ToUnshare() NLIB_NOEXCEPT {
        if (!value_) return false;
        if (nlib_atomic_sub_fetch32(value_, 1, NLIB_ATOMIC_ACQUIRE) == 0) {
            delete value_;
            value_ = NULL;
            return false;
        }
        value_ = NULL;
        return true;
    }
    bool IsUnique() NLIB_NOEXCEPT {
        if (!value_) return true;
        return nlib_atomic_load32(value_, NLIB_ATOMIC_ACQUIRE) == 1;
    }
    void swap(LspStatus& rhs) NLIB_NOEXCEPT {  // NOLINT
        using std::swap;
        swap(value_, rhs.value_);
    }

 private:
    int32_t* value_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(LspStatus);
};

}  // namespace internal

// Use of shared_ptr is discouraged because of unexpected overhead,
// and You have to share the pointer by calling MakeSharedFrom() method.
template <class T, class DEL = DefaultDelete<T> >
class LimitedSharedPtr {
 public:
    LimitedSharedPtr() NLIB_NOEXCEPT : ptr_(NULL) {}
    ~LimitedSharedPtr() NLIB_NOEXCEPT { release(); }
    explicit LimitedSharedPtr(nullptr_t) NLIB_NOEXCEPT : ptr_(NULL) {}
    explicit LimitedSharedPtr(T* p) NLIB_NOEXCEPT : ptr_(p) {
        // Do not count up LspStatus because there is the risk of alloc failure.
    }
    NLIB_MOVE_MEMBER_HELPER_2(LimitedSharedPtr, ptr_, status_)
    void reset(T* p = 0) NLIB_NOEXCEPT { LimitedSharedPtr<T>(p).swap(*this); }
    T& operator*() const {
        NLIB_ASSERT(ptr_ != NULL);
        return *ptr_;
    }
    T* operator->() const NLIB_NOEXCEPT {
        NLIB_ASSERT(ptr_ != NULL);
        return ptr_;
    }
    T* get() const NLIB_NOEXCEPT { return ptr_; }
    errno_t MakeSharedFrom(const LimitedSharedPtr& p) NLIB_NOEXCEPT;  // NOLINT
    bool IsUnique() const NLIB_NOEXCEPT { return status_.IsUnique(); }
    bool MakeUniquePtr(UniquePtr<T>& p) NLIB_NOEXCEPT;  // NOLINT
    void swap(LimitedSharedPtr& rhs) NLIB_NOEXCEPT {    // NOLINT
        using std::swap;
        swap(ptr_, rhs.ptr_);
        status_.swap(rhs.status_);
    }
    LimitedSharedPtr& operator=(nullptr_t) NLIB_NOEXCEPT {
        release();
        return *this;
    }
    NLIB_SAFE_BOOL(LimitedSharedPtr, ptr_ != NULL);

 private:
    void release() NLIB_NOEXCEPT {
        if (!status_.ToUnshare()) DEL()(ptr_);
        ptr_ = NULL;
    }

 private:
    T* ptr_;
    mutable internal::LspStatus status_;
    //    NLIB_DISALLOW_COPY_AND_ASSIGN(LimitedSharedPtr);
};

template <class T, class DEL>
errno_t LimitedSharedPtr<T, DEL>::MakeSharedFrom(const LimitedSharedPtr& p)
    NLIB_NOEXCEPT {  // NOLINT
    LimitedSharedPtr tmp;
    if (p) {
        errno_t e = p.status_.ToShare(&tmp.status_);
        if (e != 0) return e;
        tmp.ptr_ = p.ptr_;
    }
    tmp.swap(*this);
    return 0;
}

template <class T, class DEL>
bool LimitedSharedPtr<T, DEL>::MakeUniquePtr(UniquePtr<T>& p) NLIB_NOEXCEPT {  // NOLINT
    if (IsUnique()) {
        p.reset(ptr_);
        ptr_ = NULL;
        release();
        return true;
    } else {
        return false;
    }
}

template <class T, class DEL>
class LimitedSharedPtr<T[], DEL> {
 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(LimitedSharedPtr);
};  // not allowed for now

}  // namespace threading
NLIB_NAMESPACE_END

#ifndef NLIB_STD_SWAP_WORKAROUND
NLIB_NAMESPACE_BEGIN
#endif

NLIB_DEFINE_STD_SWAP_T_BEGIN1(threading)  // NOLINT
NLIB_DEFINE_STD_SWAP_T2(T, DEL, NLIB_NS::threading::LimitedSharedPtr)  // NOLINT
NLIB_DEFINE_STD_SWAP_T_END1(threading)               // NOLINT

#ifndef NLIB_STD_SWAP_WORKAROUND
NLIB_NAMESPACE_END
#endif

#endif  // INCLUDE_NN_NLIB_THREADING_LIMITEDSHAREDPTR_H_
