﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#include "xmlrpc_value.h"  // NOLINT
#include "nn/nlib/Cstring.h"
#include "nn/nlib/MemoryInputStream.h"
#include "nn/nlib/MemoryOutputStream.h"

using nlib_ns::MemoryInputStream;
using nlib_ns::MemoryOutputStream;
using nlib_ns::exi::ExiAllocator;
using nlib_ns::exi::XmlStreamReader;
using nlib_ns::exi::XmlStreamReaderSettings;
using nlib_ns::exi::XmlStreamWriter;
using nlib_ns::exi::XmlStreamWriterSettings;
using nlib_ns::exi::ExiChar;
using nlib_ns::SnPrintf;
using nlib_ns::StrCmp;

void XmlRpcValue::SetArray() {
    this->Invalidate();
    value_type_ = ARRAY;
    value_holder_.value_array = new (std::nothrow) XmlRpcArray();
}

void XmlRpcValue::SetStruct() {
    this->Invalidate();
    value_type_ = STRUCT;
    value_holder_.value_struct = new (std::nothrow) XmlRpcStruct();
}

void XmlRpcValue::Invalidate() {
    switch (value_type_) {
    case STRING:
        delete value_holder_.value_string;
        break;
    case ARRAY:
        delete value_holder_.value_array;
        break;
    case STRUCT:
        delete value_holder_.value_struct;
        break;
    default:
        break;
    }
    value_type_ = NONE;
    value_holder_.value_int = 0;
}

XmlRpcArray::~XmlRpcArray() {
    size_t i;
    size_t n = array_value_.size();
    for (i = 0; i < n; ++i) {
        delete array_value_[i];
    }
}

XmlRpcStruct::~XmlRpcStruct() {
    StructValue::iterator it;
    StructValue::iterator it_end = struct_value_.end();
    for (it = struct_value_.begin(); it != it_end; ++it) {
        delete it->second;
    }
}

XmlRpcValue& XmlRpcStruct::operator[](const StdString& rhs) {
    XmlRpcValue*& v = struct_value_[rhs];
    if (v == NULL) {
        v = new (std::nothrow) XmlRpcValue();
        NLIB_ASSERT(v != NULL);
    }
    return *v;
}

void XmlRpcValue::Write(XmlStreamWriter* writer) const {
#ifdef NLIB_EXI_WCHAR_INTERFACE
    wchar_t buf[128];
#endif
    char b[128];
    writer->WriteStartElement(N("value"));

    switch (value_type_) {
    case INT:
        writer->WriteStartElement(N("i4"));
        SnPrintf(b, "%d", value_holder_.value_int);
#ifdef NLIB_EXI_WCHAR_INTERFACE
        if (!nlib_is_error(nlib_utf8_to_wide(NULL, buf, &b[0]))) writer->WriteCharacters(buf);
#else
        writer->WriteCharacters(b);
#endif
        break;
    case BOOLEAN:
        writer->WriteStartElement(N("boolean"));
        writer->WriteCharacters(value_holder_.value_bool ? N("1") : N("0"));
        break;
    case DOUBLE:
        writer->WriteStartElement(N("double"));
        SnPrintf(b, "%f", value_holder_.value_double);
#ifdef NLIB_EXI_WCHAR_INTERFACE
        if (!nlib_is_error(nlib_utf8_to_wide(NULL, buf, &b[0]))) writer->WriteCharacters(buf);
#else
        writer->WriteCharacters(b);
#endif
        break;
    case STRING:
        writer->WriteStartElement(N("string"));
        writer->WriteCharacters(value_holder_.value_string->c_str());
        break;
    case NIL:
        writer->WriteStartElement(N("nil"));
        break;
    case ARRAY:
        writer->WriteStartElement(N("array"));
        writer->WriteStartElement(N("data"));
        value_holder_.value_array->Write(writer);
        writer->WriteEndElement();
        break;
    case STRUCT:
        writer->WriteStartElement(N("struct"));
        value_holder_.value_struct->Write(writer);
        break;
    default:
        break;
    }

    writer->WriteEndElement();
    writer->WriteEndElement();
}

void XmlRpcArray::Write(XmlStreamWriter* writer) const {
    size_t n = array_value_.size();
    size_t i;

    for (i = 0; i < n; ++i) {
        array_value_[i]->Write(writer);
    }
}

void XmlRpcStruct::Write(XmlStreamWriter* writer) const {
    StructValue::const_iterator it;
    StructValue::const_iterator it_end = struct_value_.end();

    for (it = struct_value_.begin(); it != it_end; ++it) {
        writer->WriteStartElement(N("member"));
        writer->WriteStartElement(N("name"));
        writer->WriteCharacters(it->first.c_str());
        writer->WriteEndElement();

        it->second->Write(writer);
        writer->WriteEndElement();
    }
}

namespace {

inline bool ReadEvent(XmlStreamReader* reader, XmlStreamReader::XmlStreamConstants e) {
    return reader->Next() == e;
}

#define SE \
    if (!ReadEvent(reader, XmlStreamReader::START_ELEMENT)) return false
#define EE \
    if (!ReadEvent(reader, XmlStreamReader::END_ELEMENT)) return false
#define CH \
    if (!ReadEvent(reader, XmlStreamReader::CHARACTERS)) return false

bool CheckSE(XmlStreamReader* reader, const ExiChar* name) {
    XmlStreamReader::XmlStreamConstants e = reader->Next();
    if (e != XmlStreamReader::START_ELEMENT) return false;
    return StrCmp(name, reader->GetLocalName()) == 0;
}

bool CheckEE(XmlStreamReader* reader, const ExiChar* /* name */) {
    XmlStreamReader::XmlStreamConstants e = reader->Next();
    if (e != XmlStreamReader::END_ELEMENT) return false;
    return true;
}

#define CHK_SE(name) \
    if (!CheckSE(reader, name)) return false
#define CHK_EE(name) \
    if (!CheckEE(reader, name)) return false

}  // namespace

bool XmlRpcValue::Read_(XmlStreamReader* reader) {
    if (StrCmp(reader->GetLocalName(), N("value")) != 0) return false;
    SE;

    const ExiChar* e = reader->GetLocalName();
    switch (e[0]) {
    case 'a':
        if (StrCmp(e, N("array")) == 0) {
            this->SetArray();
            if (!this->AsArray()->Read(reader)) return false;
            CHK_EE(N("array"));
        }
        break;
    case 'i':
        if (StrCmp(e, N("i4")) == 0) {
            int ival;
            CH;
            errno_t err = nlib_strto_int32(&ival, M(reader->GetText()), NULL, 10);
            if (nlib_is_error(err)) ival = 0;
            this->SetInt(ival);
            CHK_EE(N("i4"));
        } else if (StrCmp(e, N("int")) == 0) {
            int ival;
            CH;
            errno_t err = nlib_strto_int32(&ival, M(reader->GetText()), NULL, 10);
            if (nlib_is_error(err)) ival = 0;
            this->SetInt(ival);
            CHK_EE(N("int"));
        }
        break;
    case 'b':
        if (StrCmp(e, N("boolean")) == 0) {
            int ival;
            CH;
            errno_t err = nlib_strto_int32(&ival, M(reader->GetText()), NULL, 10);
            if (nlib_is_error(err)) ival = 0;
            this->SetBool(ival != 0);
            CHK_EE(N("boolean"));
        }
        break;
    case 's':
        if (StrCmp(e, N("string")) == 0) {
            CH;
            this->SetString(reader->GetText());
            CHK_EE(N("string"));
        } else if (StrCmp(e, N("struct")) == 0) {
            this->SetStruct();
            if (!this->AsStruct()->Read(reader)) return false;
            // </struct> is already gotten.
        }
        break;
    case 'd':
        if (StrCmp(e, N("double")) == 0) {
            double fval;
            CH;
            errno_t err = nlib_strto_double(&fval, M(reader->GetText()), NULL);
            if (nlib_is_error(err)) fval = 0.0;
            this->SetDouble(fval);
            CHK_EE(N("double"));
        }
        break;
    default:
        break;
    }
    CHK_EE(N("value"));
    return true;
}

bool XmlRpcArray::Read(XmlStreamReader* reader) {
    CHK_SE(N("data"));

    XmlStreamReader::XmlStreamConstants e;
    while ((e = reader->Next()) == XmlStreamReader::START_ELEMENT) {
        if (!this->Append()->Read_(reader)) return false;
    }

    // This assumes </data> if it encounters END_ELEMENT.
    if (e != XmlStreamReader::END_ELEMENT) return false;
    return true;
}

bool XmlRpcStruct::Read(XmlStreamReader* reader) {
    XmlStreamReader::XmlStreamConstants e;
    while ((e = reader->Next()) == XmlStreamReader::START_ELEMENT) {
        if (StrCmp(N("member"), reader->GetLocalName()) != 0) return false;

        CHK_SE(N("name"));
        CH;
        const ExiChar* name = reader->GetText();
        CHK_EE(N("name"));

        if (!this->Insert(name)->Read(reader)) return false;

        CHK_EE(N("member"));
    }
    if (e != XmlStreamReader::END_ELEMENT || StrCmp(N("struct"), reader->GetLocalName()) != 0)
        return false;
    return true;
}
